import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.Arrays;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.GenericWritable;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.io.Writable;
import org.apache.hadoop.io.WritableComparable;
import org.apache.hadoop.io.WritableComparator;
import org.apache.hadoop.mapreduce.InputSplit;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.Mapper;
import org.apache.hadoop.mapreduce.Partitioner;
import org.apache.hadoop.mapreduce.RecordReader;
import org.apache.hadoop.mapreduce.Reducer;
import org.apache.hadoop.mapreduce.TaskAttemptContext;
import org.apache.hadoop.mapreduce.lib.input.FileInputFormat;
import org.apache.hadoop.mapreduce.lib.input.LineRecordReader;
import org.apache.hadoop.mapreduce.lib.input.MultipleInputs;
import org.apache.hadoop.mapreduce.lib.output.FileOutputFormat;
import org.apache.hadoop.mapreduce.lib.partition.HashPartitioner;

public class SecondarySortTest {

    public static class SecondarySortableTextKey implements WritableComparable<SecondarySortableTextKey> {
        private Text textKey;
        private IntWritable secondarySortOrder;

        public SecondarySortableTextKey() {
        }

        public SecondarySortableTextKey(Text textKey, IntWritable secondarySortOrder) {
            this.textKey = textKey;
            this.secondarySortOrder = secondarySortOrder;
        }

        public Text getTextKey() {
            return this.textKey;
        }

        @Override
        public void write(DataOutput out) throws IOException {
            textKey.write(out);
            secondarySortOrder.write(out);
        }

        @Override
        public void readFields(DataInput in) throws IOException {
            Text textKey = new Text();
            textKey.readFields(in);
            IntWritable secondarySortOrder = new IntWritable();
            secondarySortOrder.readFields(in);

            this.textKey = textKey;
            this.secondarySortOrder = secondarySortOrder;
        }

        @Override
        public int compareTo(SecondarySortableTextKey secondarySortableTextKey) {
            int comparisonResult = textKey.compareTo(secondarySortableTextKey.textKey);
            if (0 == comparisonResult) {
                comparisonResult = secondarySortOrder.compareTo(secondarySortableTextKey.secondarySortOrder);
            }
            return comparisonResult;
        }
    }

    public static class SecondarySortableTextKeyPartitioner extends Partitioner<SecondarySortableTextKey, Writable> {
        private static final HashPartitioner<Text, Writable> HASH_PARTITIONER = new HashPartitioner<Text, Writable>();

        @Override
        public int getPartition(SecondarySortableTextKey key, Writable value, int numPartitions) {
            return HASH_PARTITIONER.getPartition(key.getTextKey(), value, numPartitions);
        }
    }

    public static class SecondarySortableTextKeySortComparator extends WritableComparator {
        protected SecondarySortableTextKeySortComparator() {
            super(SecondarySortableTextKey.class, true);
        }

        @SuppressWarnings("rawtypes")
        @Override
        public int compare(WritableComparable w1, WritableComparable w2) {
            SecondarySortableTextKey k1 = (SecondarySortableTextKey) w1;
            SecondarySortableTextKey k2 = (SecondarySortableTextKey) w2;

            int compare = k1.textKey.compareTo(k2.textKey);
            if (compare == 0) {
                compare = k1.secondarySortOrder.compareTo(k2.secondarySortOrder);
            }
            return compare;
        }
    }

    public static class SecondarySortableTextKeyGroupingComparator extends WritableComparator {
        protected SecondarySortableTextKeyGroupingComparator() {
            super(SecondarySortableTextKey.class, true);
        }

        @SuppressWarnings("rawtypes")
        @Override
        public int compare(WritableComparable w1, WritableComparable w2) {
            SecondarySortableTextKey k1 = (SecondarySortableTextKey) w1;
            SecondarySortableTextKey k2 = (SecondarySortableTextKey) w2;

            return k1.textKey.compareTo(k2.textKey);
        }
    }

    public static class AttributeMap extends Mapper<Text, Attribute, SecondarySortableTextKey, MyGenericWritable> {
        public void map(Text key, Attribute value, Context context) throws IOException, InterruptedException {
            System.out.println("AttributeMap:" + key.toString() + " " + value.toString());
            context.write(new SecondarySortableTextKey(key, new IntWritable(0)), new MyGenericWritable(value));
        }
    }

    public static class DataMap extends Mapper<Text, Data, SecondarySortableTextKey, MyGenericWritable> {
        public void map(Text key, Data value, Context context) throws IOException, InterruptedException {
            System.out.println("DataMap:" + key.toString() + " " + value.toString());
            context.write(new SecondarySortableTextKey(key, new IntWritable(1)), new MyGenericWritable(value));
        }
    }

    public static class MyReducer extends Reducer<SecondarySortableTextKey, MyGenericWritable, Text, Text> {
        public void reduce(SecondarySortableTextKey key, Iterable<MyGenericWritable> values, Context context)
                throws IOException, InterruptedException {
            Attribute attribute = null;
            for (MyGenericWritable value : values) {
                Writable rawValue = value.get();
                if (rawValue instanceof Attribute) {
                    System.out.println(key.getTextKey() + " Attribute");
                    attribute = (Attribute) rawValue;
                }
                if (rawValue instanceof Data) {
                    System.out.println(key.getTextKey() + " SecondClass");
                    Data data = (Data) rawValue;
                    String result;
                    if (attribute == null) {
                        result = data.data;
                    } else {
                        result = data.data + "\t" + attribute.attr1 + "\t" + attribute.attr2;
                    }
                    context.write(key.getTextKey(), new Text(result));
                }
            }
        }
    }

    public static class Attribute implements Writable {
        private String key;
        private String attr1, attr2;

        public Attribute() {
        }

        public Attribute(String key, String attr1, String attr2) {
            this.key = key;
            this.attr1 = attr1;
            this.attr2 = attr2;
        }

        @Override
        public void readFields(DataInput in) throws IOException {
            if (null == in) {
                throw new IllegalArgumentException("in cannot be null");
            }
            String key = in.readUTF();
            String attr1 = in.readUTF();
            String attr2 = in.readUTF();

            this.key = key;
            this.attr1 = attr1;
            this.attr2 = attr2;
        }

        @Override
        public void write(DataOutput out) throws IOException {
            if (null == out) {
                throw new IllegalArgumentException("out cannot be null");
            }
            out.writeUTF(this.key);
            out.writeUTF(this.attr1);
            out.writeUTF(this.attr2);
        }

        @Override
        public String toString() {
            return "Attribute\t" + key + "\t" + attr1 + "\t" + attr2;
        }

    }

    public static class Data implements Writable {

        private String key;
        private String data;

        public Data() {
        }

        public Data(String key, String data) {
            this.key = key;
            this.data = data;
        }

        @Override
        public void readFields(DataInput in) throws IOException {
            if (null == in) {
                throw new IllegalArgumentException("in cannot be null");
            }
            String key = in.readUTF();
            String data = in.readUTF();

            this.key = key;
            this.data = data;
        }

        @Override
        public void write(DataOutput out) throws IOException {
            if (null == out) {
                throw new IllegalArgumentException("out cannot be null");
            }
            out.writeUTF(this.key);
            out.writeUTF(this.data);
        }

        @Override
        public String toString() {
            return "Data\t" + key + "\t" + data;
        }
    }

    public static class AttributeReader extends RecordReader<Text, Attribute> {
        private LineRecordReader lineRecordReader = null;
        private Text key = null;
        private Attribute attribute = null;

        @Override
        public void close() throws IOException {
            if (null != lineRecordReader) {
                lineRecordReader.close();
                lineRecordReader = null;
            }
            key = null;
            attribute = null;
        }

        @Override
        public Text getCurrentKey() throws IOException, InterruptedException {
            return key;
        }

        @Override
        public Attribute getCurrentValue() throws IOException, InterruptedException {
            return attribute;
        }

        @Override
        public float getProgress() throws IOException, InterruptedException {
            return lineRecordReader.getProgress();
        }

        @Override
        public void initialize(InputSplit split, TaskAttemptContext context) throws IOException, InterruptedException {
            close();

            lineRecordReader = new LineRecordReader();
            lineRecordReader.initialize(split, context);
        }

        @Override
        public boolean nextKeyValue() throws IOException, InterruptedException {
            if (!lineRecordReader.nextKeyValue()) {
                key = null;
                attribute = null;
                return false;
            }

            // otherwise, take the line and parse it
            Text line = lineRecordReader.getCurrentValue();
            String str = line.toString();
            System.out.println("Attribute:" + str);
            String[] arr = str.split("\t", -1);

            key = new Text(arr[0].trim());
            attribute = new Attribute(arr[0].trim(), arr[1].trim(), arr[2].trim());

            return true;
        }

    }

    public static class DataReader extends RecordReader<Text, Data> {

        private LineRecordReader lineRecordReader = null;
        private Text key = null;
        private Data data = null;

        @Override
        public void close() throws IOException {
            if (null != lineRecordReader) {
                lineRecordReader.close();
                lineRecordReader = null;
            }
            key = null;
            data = null;
        }

        @Override
        public Text getCurrentKey() throws IOException, InterruptedException {
            return key;
        }

        @Override
        public Data getCurrentValue() throws IOException, InterruptedException {
            return data;
        }

        @Override
        public float getProgress() throws IOException, InterruptedException {
            return lineRecordReader.getProgress();
        }

        @Override
        public void initialize(InputSplit split, TaskAttemptContext context) throws IOException, InterruptedException {
            close();

            lineRecordReader = new LineRecordReader();
            lineRecordReader.initialize(split, context);
        }

        @Override
        public boolean nextKeyValue() throws IOException, InterruptedException {
            if (!lineRecordReader.nextKeyValue()) {
                key = null;
                data = null;
                return false;
            }

            // otherwise, take the line and parse it
            Text line = lineRecordReader.getCurrentValue();

            String str = line.toString();

            System.out.println("SecondClass:" + str);
            String[] arr = str.split("\t", -1);

            key = new Text(arr[0].trim());
            data = new Data(arr[0].trim(), arr[1].trim());

            return true;
        }

    }

    public static class AttributeInputFormat extends FileInputFormat<Text, Attribute> {
        @Override
        public RecordReader<Text, Attribute> createRecordReader(InputSplit split, TaskAttemptContext context)
                throws IOException, InterruptedException {
            return new AttributeReader();
        }
    }

    public static class DataInputFormat extends FileInputFormat<Text, Data> {
        @Override
        public RecordReader<Text, Data> createRecordReader(InputSplit split, TaskAttemptContext context)
                throws IOException, InterruptedException {
            return new DataReader();
        }
    }

    public static class MyGenericWritable extends GenericWritable {

        @SuppressWarnings("unchecked")
        private static Class<? extends Writable>[] CLASSES = (Class<? extends Writable>[]) new Class[] {
                Attribute.class, Data.class };

        // this empty initialize is required by Hadoop
        public MyGenericWritable() {
        }

        public MyGenericWritable(Writable instance) {
            set(instance);
        }

        @Override
        protected Class<? extends Writable>[] getTypes() {
            return CLASSES;
        }

        @Override
        public String toString() {
            return "MyGenericWritable [getTypes()=" + Arrays.toString(getTypes()) + "]";
        }
    }

    /**
     * @throws ClassNotFoundException
     * @throws InterruptedException
     * @param args
     * @throws
     */
    public static void main(String[] args) throws IOException, InterruptedException, ClassNotFoundException {
        Path firstPath = new Path(args[0]);
        Path sencondPath = new Path(args[1]);
        Path outputPath = new Path(args[2]);

        Configuration conf = new Configuration();

        Job job = new Job(conf);
        job.setJarByClass(SecondarySortTest.class);
        job.setJobName("MultipleInputs Test");

        job.setReducerClass(MyReducer.class);

        job.setMapOutputKeyClass(SecondarySortableTextKey.class);
        job.setMapOutputValueClass(MyGenericWritable.class);

        job.setOutputKeyClass(Text.class);
        job.setOutputValueClass(Text.class);

        job.setPartitionerClass(SecondarySortableTextKeyPartitioner.class);
        job.setSortComparatorClass(SecondarySortableTextKeySortComparator.class);
        job.setGroupingComparatorClass(SecondarySortableTextKeyGroupingComparator.class);

        System.out.println("start");

        MultipleInputs.addInputPath(job, firstPath, AttributeInputFormat.class, AttributeMap.class);
        MultipleInputs.addInputPath(job, sencondPath, DataInputFormat.class, DataMap.class);

        FileOutputFormat.setOutputPath(job, outputPath);

        job.waitForCompletion(true);
    }
}
